from .usefull_little_functions import get_seg_occ
from .intersect import invert_seg, Features
from .detect_copies import find_gene_copies,get_first_last_seg


# functions to find the paths of the features in a target genome

# add the paths of the feature on the target genome in the object Feature
def add_target_genome_paths(feature_id,target_genome_paths,segments_on_target_genome):
    feature=Features[feature_id]
    list_seg=feature.segments_list_source
    list_first_last_segs=get_first_last_seg(list_seg,segments_on_target_genome)

    for match in list_first_last_segs: # for each walk that has the feature
        [first_seg,last_seg,walk_name]=match
        # get the first and last segments of all the copies on this walk
        first_last_segs_list=find_gene_copies(list_seg,walk_name,feature_id,segments_on_target_genome)
        copy_number=0
        for first_seg,last_seg in first_last_segs_list: # get the feature path for all the copies
            copy_number+=1
            copy_id="copy_"+str(copy_number) # get the copy that corresponds to this pair of first_seg,last_seg
            feature_target_path=get_feature_path(target_genome_paths[walk_name],first_seg,last_seg,walk_name,copy_id,feature_id,segments_on_target_genome)
            feature_path=[walk_name,copy_id,feature_target_path] # informations about the occurence of the feature
            feature.segments_list_target.append(feature_path)

            # add the feat_copy to all the segments

            # get the pos start of the first and last segments (existing because done in find_gene_copies)
            # then look for segs that are between.
            walk_start=get_seg_occ(feature_target_path[0],walk_name,feature_id,copy_id,segments_on_target_genome)[1]
            walk_stop=get_seg_occ(feature_target_path[-1],walk_name,feature_id,copy_id,segments_on_target_genome)[2]
            feat_copy=(feature_id,copy_id)
            for seg in feature_target_path[1:-1]: # the first and last segs are already done in find_gene_copies
                for segment in segments_on_target_genome[seg][walk_name]: # find the right occurence
                    if walk_start <= segment[1] <= walk_stop: 
                        segment.append(feat_copy) # annotate the segment : feature_id, copy_id
                        break

            # add the paths of the gene's child features
            add_childs_paths(feature_id,feature_target_path,walk_name,copy_id,target_genome_paths,segments_on_target_genome)
     
    if len(list_first_last_segs)==0: # the latter steps expect this list to not be empty. ALSO DO IT FOR THE CHILDS ?
        feature.segments_list_target.append(['','',[]])

def add_childs_paths(feature_id,feature_target_path,walk_name,copy_id,target_genome_paths,segments_on_target_genome):
    childs_list=Features[feature_id].get_child_list(Features)

    # make a dictionnary with the segments list to search into.
    set_target = set(feature_target_path)

    for child in childs_list:

        # find child path in parent's path
        [child_first_seg,child_last_seg]=['','']
        for seg in child.segments_list_source:
            if seg in set_target: # parent path
                child_first_seg=seg
                break
            elif invert_seg(seg) in set_target:
                child_first_seg=invert_seg(seg)
                break

        if child_first_seg!='': # the child feature may be absent in this occurence of the parent
            for seg in reversed(child.segments_list_source):
                if seg in set_target: # parent path
                    child_last_seg=seg
                    break
                elif invert_seg(seg) in set_target:
                    child_last_seg=invert_seg(seg)
                    break
            
            # get the path of the child feature
            child_path=get_feature_path(target_genome_paths[walk_name],child_first_seg,child_last_seg,walk_name,copy_id,feature_id,segments_on_target_genome)
            feat_copy=(child.id,copy_id)

            feature_path=[walk_name,copy_id]
            feature_path.append(child_path)
            child.segments_list_target.append(feature_path)

            for segment_id in child_path: # annotate the segments with info about the occurence of the child feature
                segment=get_seg_occ(segment_id,walk_name,feature_id,copy_id,segments_on_target_genome)
                segment.append(feat_copy)


# find the feature's path in target genome walk
def get_feature_path(target_genome_path,first_seg,last_seg,walk_name,copy_id,feature_id,segments_on_target_genome):
    # look for first_seg and last_seg that has the right copy_id for this feature
    first_seg_index=get_seg_occ(first_seg,walk_name,feature_id,copy_id,segments_on_target_genome)[4]
    last_seg_index=get_seg_occ(last_seg,walk_name,feature_id,copy_id,segments_on_target_genome)[4]
    first_index=min(first_seg_index,last_seg_index)
    last_index=max(first_seg_index,last_seg_index)
    feature_path_target_genome=target_genome_path[first_index:last_index+1]
    return feature_path_target_genome
